import java.io.IOException;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.Date;

/*
 -p port for this proxy
 
 -r remote port for direct mode
 -d remote address for direct mode
 -b Bindaddress
 -m Maxclients
 -h helostring

 -s secured mode


Additions Matthias Gaertner 08/2005:

 -C catch-all e-Mail address
    This works as follows: the proxy replaces the argument of the first RCPT command
    with "RCPT To: " followd by the value of the -C argument. Following RCPT fields
    are silently ignored.
    Example use: -C mgr@some.where.local 

 -F force-From e-Mail address
    This works as follows: the proxy replaces the argument of the first MAIL command
    with "MAIL From: " followd by the value of the -F argument. Following MAIL fields
    are silently ignored (there should be none anyway).
    Example use: -F mgr@some.where.local 

 -X include "X-CatchAll-From/-To:" header to make replacements transparent. Existing
    X-CatchAll headers will not be removed.
 
 Note: -C and -F are independent from each other. You may use none, either or both.
 */

public final class smtpproxy
{
    private transient ServerSocket server;

    public static int MAXSMTPPROXYCLIENTS = 10; // prevent overload by cdrop

    // attacks

    public final static String VERSION = "0.1.1";

    public static InetAddress targetserver;

    public static String helo = null;

    public static int targetport = 25;

    public static boolean secured = false;

    public static String catchallrecipient = null;

    public static String forcefromsender = null;

    public static boolean includeXheaders = false;

    public static final void main(String argv[]) throws IOException
    {
        int port = 110;
        InetAddress adr = null;

        for (int i = 0; i < argv.length; i++)
        {
            String a;
            a = argv[i];
            if (a.startsWith("-"))
            {
                if (a.equalsIgnoreCase("-p"))
                {
                    i++;
                    port = parseNum(argv, i, port);
                    continue;
                }

                if (a.equalsIgnoreCase("-r"))
                {
                    i++;
                    targetport = parseNum(argv, i, targetport);
                    continue;
                }
                if (a.equalsIgnoreCase("-m"))
                {
                    i++;
                    MAXSMTPPROXYCLIENTS = parseNum(argv, i, MAXSMTPPROXYCLIENTS);
                }

                if (a.equalsIgnoreCase("-d"))
                {
                    i++;
                    String hname = "";
                    hname = parseStr(argv, i, hname);
                    if (hname.length() == 0)
                        continue;
                    try
                    {
                        targetserver = InetAddress.getByName(hname);
                    }
                    catch (IOException badname)
                    {
                        System.err.println("[SMTPPROXY] Host not found: "
                                + hname);
                    }
                    continue;
                }

                if (a.equalsIgnoreCase("-h"))
                {
                    i++;
                    helo = parseStr(argv, i, helo);
                    continue;
                }

                if (a.equalsIgnoreCase("-C"))
                {
                    i++;
                    catchallrecipient = parseStr(argv, i, catchallrecipient);
                    continue;
                }

                if (a.equalsIgnoreCase("-F"))
                {
                    i++;
                    forcefromsender = parseStr(argv, i, forcefromsender);
                    continue;
                }

                if (a.equalsIgnoreCase("-b"))
                {
                    i++;
                    String hname = "";
                    hname = parseStr(argv, i, hname);
                    if (hname.length() == 0)
                        continue;
                    try
                    {
                        adr = InetAddress.getByName(hname);
                    }
                    catch (IOException badname)
                    {
                        System.err
                                .println("[SMTPPROXY] BindAddress Host not found: "
                                        + hname);
                    }
                    continue;
                }

                if (a.equalsIgnoreCase("-s"))
                {
                    secured = true;
                    continue;
                }

                if (a.equalsIgnoreCase("-X"))
                {
                    includeXheaders = true;
                    continue;
                }

                System.err.println("[SMTPPROXY] Unknown switch " + a);
                continue;
            }
            System.err.println("[SMTPPROXY] Argument " + a
                    + " doesn't follows any switch.");
        }
        if (targetserver == null)
        {
            System.err.println("[SMTPPROXY] Target server is not defined.");
            return;
        }
        if (helo == null)
        {
            try
            {
                helo = InetAddress.getLocalHost().getHostName();
                // System.err.println("[SMTPPROXY] Helo defaulted to "+helo);
            }
            catch (UnknownHostException e)
            {
                System.err
                        .println("[SMTPPROXY] Can not find hostname for this machine. Using localhost.");
                helo = "localhost";
            }
        }
        new smtpproxy(port, adr).loop();
    }

    smtpproxy(int port, InetAddress adr) throws IOException
    {
        if (adr != null)

            try
            {
                server = new ServerSocket(port, 10, adr);
            }
            catch (IOException e)
            {
                System.err.println("[SMTPPROXY] Cannot bind to my port " + port
                        + "/" + adr.getHostAddress());
                throw e;
            }
        else
            try
            {
                server = new ServerSocket(port);
            }
            catch (IOException e)
            {
                System.err.println("[SMTPPROXY] Cannot bind to my port " + port
                        + "/*");
                throw e;
            }

    }

    private final void loop()
    {
        ThreadGroup clients;
        smtpproxyworker client;
        clients = new ThreadGroup("SMTPPROXY-clients");
        // clients.setDaemon(true);
        while (true)
        {
            Socket clientSocket = null;
            client = null;
            try
            {
                clientSocket = server.accept();
            }
            catch (IOException e)
            {
                System.err.println("[SMTPPROXY] Accept failed: " + e);
                continue;
            }
            if (clients.activeCount() > MAXSMTPPROXYCLIENTS)
            {
                try
                {
                    System.err.println(new Date()
                            + " [SMTPPROXY] Active clients limit ("
                            + MAXSMTPPROXYCLIENTS + ") reached, kicking off.");
                    clientSocket.close();
                }
                catch (IOException e)
                {
                }
                finally
                {
                }
                continue;
            }
            client = new smtpproxyworker(clientSocket);
            new Thread(clients, client).start();
        } /* listen */
    }

    private final static int parseNum(String argv[], int idx, int oldvalue)
    {
        if (idx >= argv.length)
            return oldvalue;
        int newval = oldvalue;
        try
        {
            newval = Integer.valueOf(argv[idx]).intValue();
        }
        catch (Exception e)
        {
            System.err.println("[SMTPPROXY] Not a number " + argv[idx]);
        }
        return newval;
    }

    private final static String parseStr(String argv[], int idx, String oldvalue)
    {
        if (idx >= argv.length)
            return oldvalue;
        if (argv[idx].startsWith("-"))
            return oldvalue;
        return argv[idx];
    }
}
