import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.DataOutputStream;
import java.io.EOFException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.Socket;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.StringTokenizer;

public final class smtpproxyworker implements Runnable
{
    private Socket client;

    private ArrayList recipients = null;

    private ArrayList senders = null;

    smtpproxyworker(Socket cl)
    {
        client = cl;
    }

    public final void run()
    {
        BufferedReader dis;
        DataOutputStream dos;
        BufferedReader sis = null;
        DataOutputStream sos = null;

        recipients = new ArrayList();
        senders = new ArrayList();

        Socket out = null;
        boolean isFirstEmptyLine = true;
        boolean isFirstRecipient = true;
        boolean isFirstFrom = true;

        try
        {
            dis = new BufferedReader(new InputStreamReader(client
                    .getInputStream()));
            dos = new DataOutputStream(new BufferedOutputStream(client
                    .getOutputStream(), 1500));
        }
        catch (IOException ioe)
        {
            close();
            return;
        }

        // INITIALIZE

        try
        {
            out = new Socket(smtpproxy.targetserver, smtpproxy.targetport);
            sis = new BufferedReader(
                    new InputStreamReader(out.getInputStream()));
            sos = new DataOutputStream(new BufferedOutputStream(out
                    .getOutputStream(), 1500));
            readReply(sis, dos);
        }
        catch (IOException grrr)
        {
            try
            {
                dos.writeBytes("421 " + smtpproxy.helo
                        + ". Proxy SMTP Service temporary unavailable.\r\n");
                dos.close();
            }
            catch (IOException grrrf)
            {
            }
            finally
            {
                close();
            }
            return;
        }

        mainloop: while (true)
        {
            try
            {
                String req, com;
                StringTokenizer st;
                req = dis.readLine();
                if (req == null)
                    break;
                st = new StringTokenizer(req);
                if (!st.hasMoreTokens())
                {
                    send("500 Syntax error - No command entered.", dos);
                    continue;
                }
                com = st.nextToken().toUpperCase();

                if (com.equals("QUIT"))
                {
                    send("QUIT", sos);
                    sos.close();
                    send("221 " + smtpproxy.helo
                            + ". SMTP Proxy connection sucessfully closed.",
                            dos);
                    close();
                    return;
                }
                if (com.equals("HELO") || com.equals("EHLO"))
                {
                    send(com + " " + smtpproxy.helo, sos);
                    readReply(sis, dos);
                    continue;
                }
                if (com.equals("MAIL"))
                {
                    if (smtpproxy.forcefromsender != null)
                    {
                        // Unfortunately, this simple repalcement drops any SIZE information.
                        // In any case, the header rewrite stuff would invalidate the SIZE
                        // information so that the whole header would need to be cached.
                        senders.add(req);
                        if (isFirstFrom)
                        {
                            send(com + " From: " + smtpproxy.forcefromsender,
                                    sos);
                            readReply(sis, dos);
                            isFirstFrom = false;
                        }
                        else
                        {
                            send("250 Sender ignored by catch-all proxy", dos);
                        }
                    }
                    else
                    {
                        send(req, sos);
                        readReply(sis, dos);
                    }
                    isFirstRecipient = true;
                    continue;
                }
                if (smtpproxy.catchallrecipient != null && com.equals("RCPT"))
                {
                    if (smtpproxy.includeXheaders)
                    {
                        recipients.add(req);
                    }
                    if (isFirstRecipient)
                    {
                        send(com + " To: " + smtpproxy.catchallrecipient, sos);
                        readReply(sis, dos);
                        isFirstRecipient = false;
                    }
                    else
                    {
                        send("250 Recipient ignored by catch-all proxy", dos);
                    }
                    continue;
                }
                if (com.equals("DATA"))
                {
                    send(req, sos);
                    String tmp;
                    tmp = sis.readLine();
                    if (tmp == null)
                        throw new EOFException();
                    dos.writeBytes(tmp);
                    dos.writeBytes("\r\n");
                    dos.flush();
                    if (tmp.length() > 3)
                        if (tmp.charAt(0) == '4' || tmp.charAt(0) == '5')
                        {
                            continue mainloop;
                        }

                    // read input until .
                    while (true)
                    {
                        tmp = dis.readLine();
                        if (tmp == null)
                            throw new EOFException();
                        if (isFirstEmptyLine
                                && smtpproxy.forcefromsender != null
                                && tmp.startsWith("From:"))
                        {
                            senders.add(tmp);
                            sos
                                    .writeBytes("From: "
                                            + smtpproxy.forcefromsender);
                            sos.writeBytes("\r\n");
                            continue;
                        }
                        else if (isFirstEmptyLine && smtpproxy.includeXheaders
                                && tmp.equals(""))
                        {
                            // Insert CatchAll headers
                            isFirstEmptyLine = false;
                            Iterator iter = senders.iterator();
                            while (iter.hasNext())
                            {
                                String element = (String) iter.next();
                                sos.writeBytes("X-CatchAll-From: " + element);
                                sos.writeBytes("\r\n");
                            }
                            iter = recipients.iterator();
                            while (iter.hasNext())
                            {
                                String element = (String) iter.next();
                                sos.writeBytes("X-CatchAll-To: " + element);
                                sos.writeBytes("\r\n");
                            }
                        }
                        sos.writeBytes(tmp);
                        sos.writeBytes("\r\n");
                        if (tmp.equals("."))
                            break;
                    }
                    sos.flush();
                    readReply(sis, dos);
                    continue;
                }
                if (com.equals("VERB"))
                {
                    send("500 Gatewaying of VERB command is not implemented.",
                            dos);
                    continue;
                }

                // default command processing
                // send request to remote server
                send(req, sos);
                // get reply
                readReply(sis, dos);

            }
            catch (IOException z)
            {
                try
                {
                    out.close();
                }
                catch (IOException z2)
                {
                }

                close();
                return;

            }
        }
    }

    private final void close()
    {
        try
        {
            client.close();
        }
        catch (IOException ignore)
        {
        }
        finally
        {
            client = null;
        }
    }

    private final static void send(String msg, DataOutputStream d)
            throws IOException
    {
        d.writeBytes(msg);
        d.writeBytes("\r\n");
        d.flush();
    }

    private final static void readReply(BufferedReader di, DataOutputStream d)
            throws IOException
    {
        String x;
        while (true)
        {
            x = di.readLine();
            if (x == null)
                throw new EOFException();
            d.writeBytes(x);
            d.writeBytes("\r\n");
            if (x.length() > 3 && x.charAt(3) == '-')
                continue;
            break;
        }
        d.flush();
    }
}
